(* definition of type terms and pretty-printer *)
(* $Id: types.ml,v 1.13 2000/07/10 18:51:20 alamstei Exp $ *)

open Library

type tvar = Usertvar of string 
  | Systemtvar of string*int
	
type typeexp = TVar of tvar
  | Arrow of typeexp*typeexp
  | Int
  | Bool
  | Sum of typeexp list 
  | Prod of typeexp list
	
type typeschema = (tvar list) * typeexp
      
(* Some constructors *)
      
(* needed to insure that you always have a fresh type variable for
   unification *)
let tvar_counter = ref 0
    
(* produces a fresh tvar with a specified name part *)
let fresh_named_tvar name =
  tvar_counter := (!tvar_counter) + 1;
  Systemtvar(name,!tvar_counter)

(* produces a fresh tvar with no name part *)
let fresh_tvar() = fresh_named_tvar ""

(* given a tvar do either one of the following.  if it is a usertvar return a
   fresh system tvar with no name part.  if it is a systemtvar then return a 
   tvar with the same name but a fresh int part *)
let rename_tvar tvar =
  match tvar with
    Usertvar(name) -> fresh_named_tvar ""
  | Systemtvar(name,int) -> fresh_named_tvar name

(* return a fresh typeexp *)
let fresh_typeexp () =
  TVar((fresh_tvar()))

(* puts parenthesis around string s if bool b is true *)
let optpar b s = if b then "("^s^")" else s
    
(* take a tvar and return the string of it *)
let string_of_tvariable var =
  match var with
    Usertvar(a) -> a
  | Systemtvar(name,a) -> name ^"!"^ string_of_int a
  	
(* Pretty-printer *)
let rec string_of_typeexp_prec prec a =
  match a with
    TVar(tvar) -> (string_of_tvariable tvar)
  | Arrow(b,c) -> optpar (prec>=10) ((string_of_typeexp_prec 10 b) ^ 
				     " -> " ^ (string_of_typeexp_prec 0 c))
  | Int -> "int"
  | Bool -> "bool"
  | Sum([]) -> "0"
  | Prod([]) -> "1"
  | Sum([t]) -> "(+ "^ string_of_typeexp_prec 0 t ^")"
  | Prod([t]) -> "(* "^ string_of_typeexp_prec 0 t ^")"
  | Sum(tlist) -> 
      optpar (prec>=20) (Library.string_of_list 
	(string_of_typeexp_prec 20) "+" tlist)
  | Prod(tlist) ->
      optpar (prec>=20) (Library.string_of_list 
	(string_of_typeexp_prec 20) "*" tlist)

let string_of_typeexp a = string_of_typeexp_prec 0 a

let string_of_typeschema (q,a) =
  if q=[] then
    string_of_typeexp a
  else
    "forall " ^ (Library.string_of_list string_of_tvariable " " q) ^
    "." ^ string_of_typeexp a

(* return the free tvariables of the typeexp express - i.e. all variables
   in the typeexp express *)
let rec free_tvars_of_typeexp expression =
  match expression with
    TVar(t) -> Library.singleton t
  | Arrow(t1,t2) ->
      Library.set_union (free_tvars_of_typeexp t1) (free_tvars_of_typeexp t2)
  | Int -> []
  | Bool -> []
  | Sum(tlist) -> Library.set_flatten 
	(List.map (fun x -> free_tvars_of_typeexp x) tlist)
  | Prod(tlist) -> Library.set_flatten
	(List.map (fun x -> free_tvars_of_typeexp x) tlist)

(* return the free type variables of a typeschema - i.e. all tvars in
   the typeexp a that are not in the tvar list q *)
let rec free_tvars_of_schema (q,a) =
  Library.set_difference (free_tvars_of_typeexp a) q

(* return the free type variables of the context contxt *)
let rec free_tvars_of_context contxt =
  match contxt with
    (var,schema)::tl ->
      let free_schema_tvars = free_tvars_of_schema schema in
      Library.set_union free_schema_tvars (free_tvars_of_context tl)
  | [] -> []
