(* file typing.ml - contains type inference algorithm *)
(* $Id: typing.ml,v 1.38 2000/07/10 18:51:20 alamstei Exp $ *)

open Lambda
open Types
open Unifier

type context = (variable * typeschema) list
type judgment = context * lambdaterm * typeexp

(* ---------------------------------------------------------------------- *)
(* functions for recording type information, and for formating it for
   output at the end. The reason type information is recorded in a
   global variable, rather than printed immediately, is that the type
   of a term may change during later stages of type inference. *)

(* global variable to hold the type assertions to be printed. This is
   a list of quadruples of the form (x,s,n,m), where x is a
   Lambda.variable, s is a type schema, n>=0 is the indentation level,
   and m is 0 for a let-binding and 1 for other bindings *)
let scheduled_output = ref []

(* remember the "grand total" substitution here, for purposes of
   printing types *)
let grand_sigma = ref []

(* schedule a type assertion for output. For convenience, we pass a
   context gamma instead of the indentation level - n is just the
   length of gamma *)
let schedule_for_output x schema gamma m =
  let n = List.length gamma in
  scheduled_output := (x,schema,n,m) :: !scheduled_output

(* format type information for output. This is called by ppl.ml.
   Possible modes are: "none": return empty string. "all": give info
   for all bindings.  "let": let-bindings only. "top": top-level
   let-bindings only. Or an integer indentation level. *)

let string_of_typeinfo mode =
  let (depth,binding) =
    try 
      (int_of_string mode,1)
    with
      Failure _ ->
	match mode with
	| "none" -> (0,0)
	| "all" -> (-1,1)
	| "let" -> (-1,0)
	| "top" -> (1,0)
	| _ -> (0,0) 
  in
  let rec aux = function
    | [] -> ""
    | (x,schema,n,m)::tl ->
	if m<=binding && (n<depth || depth<0) then
	  let schema' = apply_schema !grand_sigma schema in
	  aux tl ^ 
	  (String.make n ' ') ^ 
	  Lambda.string_of_variable x ^" : "^ 
	  Types.string_of_typeschema schema' ^"\n"
	else
	  aux tl 
  in
  aux !scheduled_output
	  
(* ---------------------------------------------------------------------- *)
(* returns the most general generic instance of a schema *)
let most_general_generic_instance_of schema =
  let (tvars,a) = schema in
  let subst = List.map (fun x -> (x,TVar(rename_tvar x))) tvars in
  apply_bar subst a

(* Given a typing judgment (gamma,lambdaterm,c), return the most general
   substitution, if any, which makes this typing judgment well-typed. 
   If no such substitution exists, raise Exn.Can't_unify. *)

let rec typeinfer (gamma,lambdaterm,c) = 
  (* typeinfer_tuple gamma (m1, ..., mn) (t1, ..., tn): return the most 
     general type substitution sigma which makes each judgment (sigma gamma, 
     mi, sigma ti) well-typed.  Raise Exn.Compiler_exception 
     "Typing.typeinfer_tuple" if lengths of termlist and typelist are not 
     equal.  (this should never happen) *)

  let rec typeinfer_tuple gamma termlist typelist =
    match (termlist,typelist) with
      (hd::tl, h::t) ->
	let sigma = typeinfer (gamma,hd,h) in
	let gamma' = apply_ctxt sigma gamma in
	let t' = List.map (apply_bar sigma) t in
	let tau = typeinfer_tuple gamma' tl t' in
	compose tau sigma
    | ([], []) -> []
    | _ -> raise (Exn.Compiler_exception ("Typing.typeinfer_tuple"))
  in
  (* typeinfer_case gamma tlist c injlist: return the most general 
     substitution which makes each judgment gamma, x:a_j |- P:c well-typed, 
     for each (j,n,x,P) in the injlist.  Here a_j is the jth element of
     tlist. *)

  let rec typeinfer_case gamma tlist c injlist =
    match injlist with
      [] -> []
    | (j,n,x,p)::tl ->
	let sigma = typeinfer ((x,([],List.nth tlist (j-1)))::gamma,p,c) in
	let gamma' = apply_ctxt sigma gamma in
	let tlist' = List.map (apply_bar sigma) tlist in
	let c' = apply_bar sigma c in
	let tau = typeinfer_case gamma' tlist' c' tl in
	compose tau sigma
  in

  (* create a list of fresh typeexps *)
  
  let fresh_typeexp_list n =
    let tlist  = Library.create n in
    (List.map (fun x -> (fresh_typeexp())) tlist)
  in
  match lambdaterm with
    Var(x) -> 
      begin
      	try
	  let schema = List.assoc x gamma in
	  let a = most_general_generic_instance_of schema in
	  unify c a
      	with
	  Not_found -> 
	    raise (Exn.Unbound_variable(x))
      end
  | Abs(x,m) -> 
      let a = fresh_typeexp() 
      and b = fresh_typeexp() 
      in
      let sigma = unify c (Arrow(a,b)) in
      let gamma' = apply_ctxt sigma gamma in
      let a' = apply_bar sigma a 
      and b' = apply_bar sigma b 
      in
      schedule_for_output x ([],a') gamma' 1;
      let tau = typeinfer ((x,([],a'))::gamma', m, b') in
      compose tau sigma
  | App(m,n) ->       
      let a = fresh_typeexp() in 
      let sigma = typeinfer (gamma,m,Arrow(a,c)) in
      let gamma' = apply_ctxt sigma gamma in
      let a' = apply_bar sigma a in
      let tau = typeinfer (gamma', n, a') in
      compose tau sigma
  | Let(x,m,n) ->
      let a = fresh_typeexp() in
      let sigma = typeinfer (gamma, m, a) in
      let a' = apply_bar sigma a in
      let c' = apply_bar sigma c in
      let gamma' = apply_ctxt sigma gamma in
      let free_a' = free_tvars_of_typeexp a' in
      let free_gamma' = free_tvars_of_context gamma' in
      let q = Library.set_difference free_a' free_gamma' in
      schedule_for_output x (q,a') gamma' 0;
      let tau = typeinfer ((x,(q,a'))::gamma',n,c') in
      compose tau sigma
  | Letrec(x,m,n) ->
      let a = fresh_typeexp() in
      let sigma = typeinfer ((x,([],a))::gamma, m, a) in
      let a' = apply_bar sigma a in
      let c' = apply_bar sigma c in
      let gamma' = apply_ctxt sigma gamma in
      let free_a' = free_tvars_of_typeexp a' in
      let free_gamma' = free_tvars_of_context gamma' in
      let q = Library.set_difference free_a' free_gamma' in
      schedule_for_output x (q,a') gamma' 0;
      let tau = typeinfer ((x,(q,a'))::gamma',n,c') in
      compose tau sigma
  | Intlit(n) -> unify c Int
  | Boollit(b) -> unify c Bool
  | Tuple(tuplelist) ->
      let freshtlist = fresh_typeexp_list (List.length tuplelist) in
      let sigma = unify c (Prod(freshtlist)) in
      let typelist = List.map (apply_bar sigma) freshtlist in
      let gamma' = apply_ctxt sigma gamma in
      let tau = typeinfer_tuple gamma' tuplelist typelist in
      compose tau sigma
  | Proj(i,j,m) ->
      let typelist = fresh_typeexp_list j in 
      let sigma = unify c (List.nth typelist (i-1)) in
      let typelist' = List.map (apply_bar sigma) typelist in
      let gamma' = apply_ctxt sigma gamma in
      let tau = typeinfer (gamma',m,Prod(typelist')) in
      compose tau sigma
  | Inj(i,n,m) ->
      let typelist = fresh_typeexp_list n in
      let sigma = unify c (Sum(typelist)) in
      let typelist' = List.map (apply_bar sigma) typelist in
      let gamma' = apply_ctxt sigma gamma in
      let tau = typeinfer (gamma',m,List.nth typelist (i-1)) in
      compose tau sigma
  | Case(m,injlist) ->
      (* first determine n.  Notice that we need to look for n in the first 
	 term inj/n, rather than counting the number of cases:  there might be
	 more or less than n cases.  The only exception is if there are 0 
	 cases - in this case we assume n=0. *)
      let n = match injlist with [] -> 0 | (j,n,x,m)::t -> n in

      (* create a list of fresh type variables A1,...,An *)
      let tlist = fresh_typeexp_list n in

      (* now find sigma such that Gamma |- M:A1+...+An *)
      let sigma = typeinfer (gamma,m,Sum(tlist)) in
      let gamma' = apply_ctxt sigma gamma in
      let tlist' = List.map (apply_bar sigma) tlist in
      let c' = apply_bar sigma c in
      
      (* now find tau such that for each (j,n,x,N) in the injlist, 
	 Gamma, x:A_j |- N:C.  This is done by an auxiliary fn. *)
      let tau = typeinfer_case gamma' tlist' c' injlist in
      compose tau sigma

let most_general_type m =
  scheduled_output := [];
  let alpha = fresh_tvar() in
  let sigma = typeinfer (Stdlib.stdcontext, m, TVar(alpha)) in
  grand_sigma := sigma;
  apply sigma alpha
